import datetime
import threading
from cachetools import TTLCache


class Cache(object):
    def __init__(self, cache_size=1000,
                 cache_ttl=datetime.timedelta(minutes=15).total_seconds(),
                 cache_on_disk=False,
                 key_cache_dir=None,
                 path_cache_dir=None):
        if cache_on_disk:
            # don't make this mandatory
            from diskcache import Cache as DiskCache
            self.keys = DiskCache(key_cache_dir)
            self.paths = DiskCache(path_cache_dir)
        else:
            self.keys = TTLCache(maxsize=cache_size, ttl=cache_ttl)
            self.paths = TTLCache(maxsize=cache_size, ttl=cache_ttl)
        self.lock = threading.Lock()
        self.cache_ttl = cache_ttl
        self.cache_on_disk = cache_on_disk

    def get(self, cache, key, *args):
        with self.lock:
            c = getattr(self, cache)
            return c[key][args]

    def set(self, cache, key, value, *args):
        with self.lock:
            c = getattr(self, cache)
            try:
                c[key][args] = value
            except KeyError:
                if self.cache_on_disk:
                    # with DiskCache, we have to set the TTL on the keys
                    c.set(key, {args: value}, expire=self.cache_ttl)
                else:
                    c[key] = {args: value}

    def delete(self, cache, key, *args):
        with self.lock:
            c = getattr(self, cache)
            del c[key]

    def clear(self, cache):
        with self.lock:
            c = getattr(self, cache)
            c.clear()



from __future__ import absolute_import
import importlib


DOCSTRINGS = {
    'provider': 'provider (str): Secret provider (aws.sm|aws.ps)',
    'path': 'path (str): path for the given secret',
    'keypath': 'keypath (list): the key path for looking into a JSON secret'}


class DictQuery(dict):
    def get(self, keys, default=None):
        val = None

        for key in keys:
            if val:
                if isinstance(val, list):
                    val = [v.get(key, default) if v else None for v in val]
                else:
                    try:
                        val = val.get(key, default)
                    except AttributeError:
                        return default
            else:
                val = dict.get(self, key, default)

            if val == default:
                break
        return val


def _get_provider(provider, **params):
    m = importlib.import_module(name='.{}'.format(provider),
                                package='s1crets.providers')
    return m.SecretProvider(**params)


def docstring_parameter(sub):
    def dec(obj):
        obj.__doc__ = obj.__doc__.format(**sub)
        return obj
    return dec


@docstring_parameter(DOCSTRINGS)
def get(provider='aws.sm', path=None, keypath=None, retry=3, timeout=5):
    """Get a secret from the given `provider`

    Args:
        {provider}
        {path}
        {keypath}

    Returns:
        secret: The returned secret, can be string, bytes or in case of JSON, a dictionary
    """
    p = _get_provider(provider, retry=retry, timeout=timeout)
    return p.get(path, keypath=keypath)


@docstring_parameter(DOCSTRINGS)
def path_exists(provider='aws.sm', path=None, keypath=None, retry=3, timeout=5):
    """Check whether the path exists in the secrets provider

    Args:
        {provider}
        {path}
        {keypath}

    Returns:
        secret: The returned secret, can be string, bytes or in case of JSON, a dictionary
    """
    p = _get_provider(provider, retry=retry, timeout=timeout)
    return p.path_exists(path, keypath=keypath)


@docstring_parameter(DOCSTRINGS)
def get_by_path(provider='aws.sm', path=None, retry=3, timeout=5):
    """Returns all secrets beneath a path (if the provider supports it)

    Args:
        {provider}
        {path}

    Returns:
        secrets (list): List of returned secrets
    """
    p = _get_provider(provider, retry=retry, timeout=timeout)
    return p.get_by_path(path)


@docstring_parameter(DOCSTRINGS)
def update(provider='aws.sm', path=None, value=None, retry=3, timeout=5):
    """Updates secret with given value

    Args:
        {provider}
        {path}
        value (string, bytes): the value to be stored

    Returns:
        None
    """

    p = _get_provider(provider, retry=retry, timeout=timeout)
    return p.update(path, value)



import sys
import click
import s1crets


@click.group()
def main():
    pass


@main.command()
@click.option('--provider', help='Secrets provider', default='aws.sm',
              show_default=True, type=click.Choice(['aws.sm', 'aws.ps']))
@click.argument('path', nargs=1, required=True)
@click.argument('keypath', nargs=-1, type=click.UNPROCESSED)
def get(provider, path, keypath):
    click.echo(s1crets.get(provider, path, keypath=keypath))


@main.command()
@click.option('--provider', help='Secrets provider', default='aws.sm',
              show_default=True, type=click.Choice(['aws.sm', 'aws.ps']))
@click.argument('path', nargs=1, required=True)
@click.argument('keypath', nargs=-1, type=click.UNPROCESSED)
def exists(provider, path, keypath):
    if not s1crets.path_exists(provider, path, keypath=keypath):
        sys.exit(1)


@main.command()
@click.option('--provider', help='Secrets provider', default='aws.sm',
              show_default=True, type=click.Choice(['aws.sm', 'aws.ps']))
@click.argument('path', nargs=1, required=True)
def get_by_path(provider, path):
    for k, v in s1crets.get_by_path(provider, path).items():
        print(k, v)


@main.command()
@click.option('--provider', help='Secrets provider', default='aws.sm',
              show_default=True, type=click.Choice(['aws.sm', 'aws.ps']))
@click.argument('path', nargs=1, required=True)
@click.argument('value', nargs=1, required=True)
def update(provider, path, value):
    s1crets.update(provider, path, value)



import logging
import boto3
import requests
from botocore.client import Config
from botocore.exceptions import ClientError


INSTANCE_IDENTITY_URL = 'http://169.254.169.254/2016-09-02/dynamic/instance-identity/document'


def get_region():
    # try to get configured region from the AWS environment
    region = boto3.session.Session().region_name

    # if we don't have one, try to figure it out from the INSTANCE_IDENTITY_URL
    if region is None:
        try:
            res = requests.get(INSTANCE_IDENTITY_URL, timeout=0.2)
        except Exception:
            logging.exception(f"Couldn't get {INSTANCE_IDENTITY_URL}")
            return region
        if res.status_code == 200:
            region = res.json().get('region', region)
        else:
            logging.error(f"Got {res.status_code} for {INSTANCE_IDENTITY_URL}")
    return region


class ServiceWrapper(object):
    """Wraps an AWS service behind a temporarily held and renewed if necessary
    STS token, which makes it possible to switch roles
    """

    def __init__(self, service, RoleArn=None, RoleSessionName='s1crets',
                 DurationSeconds=3600, aws_region=None,
                 boto_config=None):
        self.service = service
        self.RoleArn = RoleArn
        self.DurationSeconds = DurationSeconds
        self.RoleSessionName = RoleSessionName
        self.session = boto3.session.Session()
        self.sts = self.session.client('sts')
        self.aws_region = aws_region or get_region()
        self.boto_config = boto_config
        self.init_service()

    def init_service(self):
        kw = {}
        if self.RoleArn is not None:
            creds = self.sts.assume_role(RoleArn=self.RoleArn,
                                         RoleSessionName=self.RoleSessionName,
                                         DurationSeconds=self.DurationSeconds)
            creds = creds['Credentials']
            kw['aws_access_key_id'] = creds['AccessKeyId']
            kw['aws_secret_access_key'] = creds['SecretAccessKey']
            kw['aws_session_token'] = creds['SessionToken']
        self.wrapped_service = self.session.client(self.service,
                                                   config=self.boto_config,
                                                   region_name=self.aws_region, **kw)

    def __getattr__(self, name):
        obj = getattr(self.wrapped_service, name)
        if not callable(obj):
            return obj
        return lambda *args, **kw: self.wrapper(name, *args, **kw)

    def wrapper(self, name, *args, **kw):
        for i in range(2):
            f = getattr(self.wrapped_service, name)
            try:
                return f(*args, **kw)
            except ClientError as e:
                # on the first ExpiredTokenException try to reinit the wrapped
                # service with a new token. The session token may have expired.
                if i == 0 and e.response['Error']['Code'] == 'ExpiredTokenException':
                    self.init_service()
                # on subsequent and other exceptions just raise
                else:
                    raise



import json
import cachetools
from botocore.client import Config
from botocore.exceptions import ClientError
from s1crets.core import DictQuery
from s1crets.providers.base import BaseProvider, DefaultValue, args_cache_key
from s1crets.providers.aws.base import ServiceWrapper


@cachetools.cached(cache={}, key=args_cache_key)
class SecretProvider(BaseProvider):
    def __init__(self, sts_args={}, cache_args={}, retry=3, timeout=5, **kwargs):
        config = Config(connect_timeout=timeout, read_timeout=timeout,
                        retries={'total_max_attempts': retry})
        self.ssm = ServiceWrapper('ssm', boto_config=config, **sts_args)
        super().__init__(sts_args=sts_args, cache_args=cache_args)

    def get(self, path, default=DefaultValue, decrypt=True, cached=True, **kwargs):
        if cached:
            try:
                # we take decrypt's value into account, so we can store both
                # encrypted and decrypted data in the cache
                return self.cache.get('keys', path, decrypt)
            except KeyError:
                # not in cache
                pass
        if not path.startswith('/aws/reference/secretsmanager/'):
            # if the path references the Parameter Store, just try to fetch
            # the value
            try:
                res = self.ssm.get_parameter(Name=path,
                                             WithDecryption=decrypt)
                self.cache.set('keys', path, res['Parameter']['Value'], decrypt)
                return res['Parameter']['Value']
            except ClientError as e:
                if e.response['Error']['Code'] == 'ParameterNotFound':
                    if default is not DefaultValue:
                        # if the parameter cannot be found and we've got a
                        # default value, return it, instead of raising
                        # "KeyError" exception
                        return default
                    else:
                        raise KeyError(path)
                # if it's not ParameterNotFound and we haven't got a default
                # value, re-raise the exception
                raise

        # Secrets Manager stores mostly a JSON, so in order to make it possible
        # to reference data stored in it in the same way as with PS, iterate
        # on the path from backwards on / separator and try to find the key
        # which we need to get
        fetch = True
        res = None
        for i, c in list(enumerate(path))[::-1]:
            if path[:i+1] == '/aws/reference/secretsmanager/':
                # don't go below the above path, there is nothing there for us
                break
            if fetch:
                try:
                    res = self.ssm.get_parameter(Name=path[:i+1],
                                                 WithDecryption=decrypt)
                    break
                except ClientError as e:
                    if e.response['Error']['Code'] != 'ParameterNotFound':
                        # let other exceptions through
                        raise
                fetch = False
            if c == '/':
                fetch = True

        # no such key
        if res is None:
            raise KeyError(path)

        try:
            # is it a JSON?
            res = json.loads(res['Parameter']['Value'])
        except Exception:
            # no
            self.cache.set('keys', path, res['Parameter']['Value'], decrypt)
            return res['Parameter']['Value']

        if not path[i+2:]:
            # if the remainder of the path is empty, the SM value was referenced
            self.cache.set('keys', path, res, decrypt)
            return res

        # otherwise a token inside the JSON was referenced, try to return that,
        # with handling nonexistent/default cases
        subkey = path[i+2:].split('/')
        if default is DefaultValue:
            res = DictQuery(res).get(subkey, DefaultValue)
            if res is DefaultValue:
                # no such key
                raise KeyError(path)
            else:
                self.cache.set('keys', path, res, decrypt)
                return res
        else:
            val = DictQuery(res).get(subkey, default)
            if val != default:
                self.cache.set('keys', path, val, decrypt)
            return val

    def get_by_path(self, path, decrypt=True, recursive=True, cached=True):
        try:
            if cached:
                return self.cache.get('paths', path, decrypt, recursive)
        except KeyError:
            # not in cache
            pass

        params = {}
        kwargs = {}
        while True:
            r = self.ssm.get_parameters_by_path(Path=path, Recursive=recursive,
                                                WithDecryption=decrypt, **kwargs)
            for param in r.get('Parameters', []):
                params[param['Name']] = param['Value']

            if 'NextToken' not in r or r['NextToken'] is None:
                # we've got all params
                break
            # set the next token
            kwargs['NextToken'] = r['NextToken']
        self.cache.set('paths', path, params, decrypt, recursive)
        return params

    def update(self, path, value):
        next_token = None
        search_next_page = True
        # the loop will stop when the response do not contain NextToken or we got the data
        while search_next_page:
            try:
                p_dict = {"Filters": [{'Key': 'Name', 'Values': [path]}]}
                if next_token:
                    p_dict['NextToken'] = next_token
                res = self.ssm.describe_parameters(**p_dict)
            except ClientError as e:
                if e.response['Error']['Code'] == 'ParameterNotFound':
                    raise KeyError(path)
                raise

            orig_params = res.get('Parameters', [])
            if not orig_params:
                if 'NextToken' not in res:
                    # can not find the path in parameter storage
                    raise KeyError(path)
                else:
                    # can not find the path in current page, need to search in another page
                    next_token = res['NextToken']
            else:
                search_next_page = False

        if len(orig_params) > 1:
            raise KeyError('describe_parameters returned other than one ({}) parameters on path {}'.format(
                len(orig_params), path))

        kwargs = self.dict_filt(orig_params[0], ('Name', 'Type', 'KeyId', 'Description'))
        self.ssm.put_parameter(Value=value, Overwrite=True, **kwargs)

        # remove path from the key_cache
        try:
            self.cache.delete('keys', path)
        except KeyError:
            pass
        # and simply drop all entries from path_cache
        self.cache.clear('paths')
        return value

    def path_exists(self, path, **kwargs):
        # we're using describe_parameters here, so we can check for paths and
        # exact keys as well
        next_token = None
        is_path_in_parameter_storage = None
        # the loop will stop when the response do not contain NextToken or we got the data
        while is_path_in_parameter_storage is None:
            try:
                p_dict = {"Filters": [{'Key': 'Name', 'Values': [path]}]}
                if next_token:
                    p_dict['NextToken'] = next_token
                res = self.ssm.describe_parameters(**p_dict)
            except ClientError as e:
                if e.response['Error']['Code'] == 'ParameterNotFound':
                    is_path_in_parameter_storage = False
                else:
                    raise

            orig_params = res.get('Parameters', [])
            # can not find it in the page
            if not orig_params:
                if 'NextToken' not in res:
                    # can not find the path in parameter storage
                    is_path_in_parameter_storage = False
                else:
                    # can not find the path in current page, need to search in another page
                    next_token = res['NextToken']
            else:
                is_path_in_parameter_storage = True
        return is_path_in_parameter_storage



import json
import cachetools
from botocore.client import Config
from s1crets.core import DictQuery
from s1crets.providers.base import BaseProvider, DefaultValue, args_cache_key
from s1crets.providers.aws.base import ServiceWrapper


@cachetools.cached(cache={}, key=args_cache_key)
class SecretProvider(BaseProvider):
    def __init__(self, sts_args={}, cache_args={}, retry=3, timeout=5, **kwargs):
        config = Config(connect_timeout=timeout, read_timeout=timeout,
                        retries={'total_max_attempts': retry})
        self.sm = ServiceWrapper('secretsmanager', boto_config=config, **sts_args)
        
        super().__init__(sts_args=sts_args, cache_args=cache_args)

    def _get_secret_value(self, secret, path, keypath=None, default=DefaultValue):
        if 'SecretBinary' in secret:
            return secret['SecretBinary']
        data = secret['SecretString']
        try:
            # secrets in Secret Manager are mostly a JSON, try to parse it
            data = json.loads(data)
        except Exception:
            pass
        else:
            if keypath:
                val = DictQuery(data).get(keypath, default)
                if val is DefaultValue:
                    raise KeyError(path, keypath)
                else:
                    return val
        return data

    def get(self, path, keypath=None, default=DefaultValue, cached=True, **kwargs):
        if cached:
            try:
                res = self.cache.get('keys', path)
            except KeyError:
                # not in cache
                pass
            else:
                return self._get_secret_value(res, path, keypath, default)

        try:
            res = self.sm.get_secret_value(SecretId=path)
            self.cache.set('keys', path, res)
        except self.sm.exceptions.ResourceNotFoundException:
            raise KeyError(path)

        return self._get_secret_value(res, path, keypath, default)

    def get_by_path(*args, **kwargs):
        # Secrets Manager doesn't support the concept of paths
        raise NotImplementedError

    def update(self, path, value):
        # get the current secret in order to see its type
        secret = self.sm.get_secret_value(SecretId=path)
        if 'SecretBinary' in secret:
            kwargs = {'SecretBinary': value}
        else:
            kwargs = {'SecretString': value}

        self.sm.put_secret_value(SecretId=path, **kwargs)
        # remove path from the key_cache
        try:
            self.cache.delete('keys', path)
        except KeyError:
            pass

    def path_exists(self, path, keypath=None, cached=True, **kwargs):
        try:
            self.get(path, keypath)
            return True
        except KeyError:
            return False



from abc import ABCMeta, abstractmethod
from s1crets.cache import Cache
import hashlib
import cloudpickle


# try to make a hashable key for caching
def args_cache_key(*args, **kw):
    return hashlib.sha256(cloudpickle.dumps([args, kw])).digest()


class DefaultValue(object):
    """Indicates a default value, which is distinguishable from None (which
    can also be a default value)
    """
    pass


class BaseProvider(object, metaclass=ABCMeta):
    def __init__(self, cache_args={}, **kwargs):
        self.cache = Cache(**cache_args)
        self.kwargs = kwargs

    @staticmethod
    def dict_filt(d, keys):
        """filter dictionary d to keys keys"""
        return dict([(i, d[i]) for i in d if i in set(keys)])

    @abstractmethod
    def get(self, path, keypath=None, default=DefaultValue, decrypt=True, cached=True,
            retry=None, timeout=None):
        pass

    @abstractmethod
    def get_by_path(self, path, decrypt=True, recursive=True, cached=True,
                    retry=None, timeout=None):
        pass

    @abstractmethod
    def update(self, path, value, retry=None, timeout=None):
        pass

    @abstractmethod
    def path_exists(self, path, keypath=None, cached=True, retry=None, timeout=None):
        pass



import io
import re
import setuptools

with io.open("README.rst", "rt", encoding="utf8") as f:
    long_description = f.read()

with io.open("s1crets/__init__.py", "rt", encoding="utf8") as f:
    version = re.search(r"__version__ = \'(.*?)\'", f.read()).group(1)

setuptools.setup(
    name="s1crets",
    version=version,
    author="NAGY, Attila",
    author_email="nagy.attila@gmail.com",
    description="Convenience wrapper for AWS Parameter Store/Secrets Manager",
    long_description=long_description,
    url="https://github.com/bra-fsn/s1crets",
    packages=setuptools.find_packages(),
    scripts=['bin/s1crets'],
    install_requires=['cachetools', 'click>=7.0', 'boto3>=1.9.23', 'requests', 'cloudpickle'],
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "Documentation": "https://s1crets.readthedocs.io/en/latest/index.html",
        "Issue tracker": "https://github.com/bra-fsn/s1crets/issues"
    }
)
